/*
 * Copyright (C): Timo Fleisch
 *                2009 Berlin, Germany
 *                http://www.vrhome.de
 *                All Rights Reserved.
 *
 * THIS SOURCECODE IS PUBLISHED UNDER RESTRICTIONS.
 * No unauthorized copy or distribution allowed.
 * Please see website for detailed liscence information.
 * 
 */
using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;

namespace Tgex
{
    public class Dynob : IRecyclable
    {
        protected Dictionary<String, ICondition> m_ConditionList = new Dictionary<string,ICondition>();
        /// <summary>
        /// Helper dictionary for building dynob.
        /// </summary>
        public Dictionary<String, ICondition> ConditionList
        {
            get { return m_ConditionList; }
            set { m_ConditionList = value; }
        }
        protected Dictionary<String, IAction> m_ActionList = new Dictionary<string,IAction>();
        /// <summary>
        /// Helper dictionary for building dynob.
        /// </summary>
        public Dictionary<String, IAction> ActionList
        {
            get { return m_ActionList; }
            set { m_ActionList = value; }
        }

        protected List<Modifier> modifierList = new List<Modifier>();
        public void AddModifier(Modifier modifier)
        {
            modifierList.Add(modifier);
            modifier.Register(this);
        }
        public Modifier GetModifierByName(String name)
        {
            foreach (Modifier modifier in modifierList)
            {
                if (modifier.Name == name)
                {
                    return modifier;
                }
            }
            return null;
        }

        protected List<Module> moduleList = new List<Module>();
        public void AddModule(Module module)
        {
            moduleList.Add(module);
            module.Register(this);
        }
        public Module GetModuleByName(String name)
        {
            foreach (Module module in moduleList)
            {
                if (module.Name == name)
                {
                    return module;
                }
            }
            return null;
        }
        protected List<Property> propertyList = new List<Property>();
        public List<Property> PropertyList
        {
            get { return propertyList; }
            set { propertyList = value; }
        }

        public void AddProperty(Property property)
        {
            propertyList.Add(property);
        }
        public Property GetPropertyByName(String name)
        {
            foreach (Property property in propertyList)
            {
                if (property.Name == name)
                {
                    return property;
                }
            }
            return GlobalProperties.Instance.GetPropertyByName(name);
        }

        protected List<Trigger> triggerList = new List<Trigger>();
        public List<Trigger> TriggerList
        {
            get { return triggerList; }
            set { triggerList = value; }
        }
        public void AddTrigger(Trigger trigger)
        {
            triggerList.Add(trigger);
        }
        public Trigger GetTriggerByName(String name)
        {
            foreach (Trigger trigger in triggerList)
            {
                if (trigger.Name == name)
                {
                    return trigger;
                }
            }
            return null;
        }

        private FloatProperty lifetime;

        public float Lifetime
        {
            get { return lifetime.Value; }
            set { lifetime.Value = value; }
        }

        private BoolProperty dead;

        public bool Dead
        {
            get { return dead.Value; }
            set { dead.Value = value; }
        }
        /// <summary>
        /// Constructor
        /// </summary>
        public Dynob()
        {
            dead = new BoolProperty();
            dead.Value = false;
            dead.Name = "dead";
            lifetime = new FloatProperty();
            lifetime.Value = 0;
            lifetime.Name = "lifetime";
            // add the property, so it can be accessed from outside
            AddProperty(dead);
            AddProperty(lifetime);
        }
        public void Initialize()
        {
            foreach (Property property in propertyList)
            {
                property.Initialize();
            }
            foreach (Module module in moduleList)
            {
                module.Initialize();
            }
            foreach (Modifier modifier in modifierList)
            {
                modifier.Initialize();
            }
        }
        public void OnStart()
        {
            foreach (Module module in moduleList)
            {
                module.OnStart();
            }
        }
        /// <summary>
        /// called in game logic to animate parameter
        /// </summary>
        /// <param name="t">Current game time</param>
        public virtual void UpdateValue(GameTime t)
        {
            if (!dead.Value)
            {
                lifetime.Value += (float)(t.ElapsedGameTime.TotalMilliseconds);
                foreach (Modifier modifier in modifierList)
                {
                    if (modifier.Active)
                    {
                        modifier.Update(t);
                    }
                }
                foreach (Module module in moduleList)
                {
                    if (module.Active)
                    {
                        module.Update(t);
                    }
                }
                foreach (Trigger trigger in TriggerList)
                {
                    if (!trigger.IsDead)
                    {
                        trigger.CheckCondition(t);
                    }
                }
            }
        }
        public void Draw(GraphicsDeviceManager graphics)
        {
            if (!dead.Value)
            {
                foreach (Module module in moduleList)
                {
                    if (module.Active)
                    {
                        module.Draw(graphics);
                    }
                }
            }
        }
        /// <summary>
        /// Set all values to initial values
        /// </summary>
        public void Reset()
        {
            dead.Value = false;
            lifetime.Value = 0;
            foreach (Property property in propertyList)
            {
                property.Reset();
            }
            foreach (Modifier modifier in modifierList)
            {
                modifier.Reset();
            }
            foreach (Module module in moduleList)
            {
                module.Reset();
            }
            foreach (Trigger trigger in TriggerList)
            {
                trigger.Reset();
            }
        }

        public void Recycle()
        {
            Reset();
        }
    }
}
