/*
 * Copyright (C): Timo Fleisch
 *                2009 Berlin, Germany
 *                http://www.vrhome.de
 *                All Rights Reserved.
 *
 * THIS SOURCECODE IS PUBLISHED UNDER RESTRICTIONS.
 * No unauthorized copy or distribution allowed.
 * Please see website for detailed liscence information.
 * 
 */
using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;

namespace Tgex.Graphics
{
    public class Camera
    {
        private Vector3 m_eyePosition;
        /// <summary>
        /// Position of camera
        /// </summary>
        public Vector3 EyePosition
        {
            get { return m_eyePosition; }
            set { m_eyePosition = value; }
        }
        private Vector3 m_upVector;
        /// <summary>
        /// Up vector of camera.
        /// </summary>
        public Vector3 UpVector
        {
            get { return m_upVector; }
            set { m_upVector = value; }
        }
        private Vector3 m_lookAt;
        /// <summary>
        /// Point the camera is looking at. In stereo mode this is the zero parallax point.
        /// </summary>
        public Vector3 LookAt
        {
            get { return m_lookAt; }
            set { m_lookAt = value; }
        }
        private float m_fieldOfView;
        /// <summary>
        /// Horizontal field of view.
        /// </summary>
        public float FieldOfView
        {
            get { return m_fieldOfView; }
            set { m_fieldOfView = value; }
        }
        private float m_aspectRatio;
        /// <summary>
        /// Aspect ratio of view view frustum
        /// </summary>
        public float AspectRatio
        {
            get { return m_aspectRatio; }
            set { m_aspectRatio = value; }
        }
        private float m_nearPlaneDistance;
        /// <summary>
        /// Distance to near clipping plane
        /// </summary>
        public float NearPlaneDistance
        {
            get { return m_nearPlaneDistance; }
            set { m_nearPlaneDistance = value; }
        }
        private float m_farPlaneDistance;
        /// <summary>
        /// Distance to far clipping plane.
        /// </summary>
        public float FarPlaneDistance
        {
            get { return m_farPlaneDistance; }
            set { m_farPlaneDistance = value; }
        }
        private Matrix m_projectionMatrix;
        /// <summary>
        /// Projection matrix
        /// </summary>
        public Matrix ProjectionMatrix
        {
            get { return m_projectionMatrix; }
            set { m_projectionMatrix = value; }
        }
        private Matrix m_viewMatrix;
        /// <summary>
        /// View matrix
        /// </summary>
        public Matrix ViewMatrix
        {
            get { return m_viewMatrix; }
            set { m_viewMatrix = value; }
        }
        private StereoViewMode m_stereoMode;
        /// <summary>
        /// Stereo mode of camera
        /// </summary>
        public StereoViewMode StereoMode
        {
            get { return m_stereoMode; }
            set { m_stereoMode = value; }
        }
        private float m_eyeDistance;
        /// <summary>
        /// Distance between left and right eye. Used for stereo camera calculation.
        /// </summary>
        public float EyeDistance
        {
            get { return m_eyeDistance; }
            set { m_eyeDistance = value; }
        }
        /// <summary>
        /// Techniques how to calculate the stereo view projection.
        /// </summary>
        public enum StereoViewProjectionMode
        {
            ToeInStereo,
            ShearedViewFrustum
        }
        private StereoViewProjectionMode m_stereoViewProjectionMode;
        /// <summary>
        /// Defines how the stereo projection view matrix is calculated.
        /// </summary>
        public StereoViewProjectionMode StereoProjectionMode
        {
            get { return m_stereoViewProjectionMode; }
            set { m_stereoViewProjectionMode = value; }
        }
        /// <summary>
        /// Defines the left and right eye.
        /// </summary>
        public enum Eye
        {
            Left,
            Right
        }
        private Eye m_currentEye;
        /// <summary>
        /// Sets and gets the current eye for stereo rendering.
        /// </summary>
        public Eye CurrentEye
        {
            get { return m_currentEye; }
            set { m_currentEye = value; }
        }
        /// <summary>
        /// Set default values
        /// </summary>
        public void Initialize()
        {
            m_eyePosition = new Vector3(0, 2, 5);
            m_lookAt = Vector3.Zero;
            m_upVector = Vector3.Up;
            m_fieldOfView = MathHelper.ToRadians(70);
            m_nearPlaneDistance = 0.001f;
            m_farPlaneDistance = 100.0f;
            m_aspectRatio = 4.0f / 3.0f;
            m_projectionMatrix = Matrix.Identity;
            m_viewMatrix = Matrix.Identity;

            // 7 cm
            m_eyeDistance = 0.07f;
            m_currentEye = Eye.Left;
            m_stereoMode = StereoViewMode.None;
            m_stereoViewProjectionMode = StereoViewProjectionMode.ToeInStereo;

            UpdateProjectionMatrix();
            UpdateViewMatrix();
        }
        /// <summary>
        /// Calculate the view matrix
        /// </summary>
        public void UpdateViewMatrix()
        {
            if (m_stereoMode == StereoViewMode.None)
            {
                m_viewMatrix = Matrix.CreateLookAt(
                    m_eyePosition,
                    m_lookAt,
                    m_upVector);
            }
            else
            {
                // temp cam paras
                Vector2 displaySize = new Vector2(0.376f, 0.303f);
                bool m_projectToScreen = true;

                Vector3 vEyePt = m_eyePosition;
                Vector3 vLookatPt = m_lookAt;
                Vector3 vUp = m_upVector;

                // calculate adjusted eye positions
                Vector3 nLook = vLookatPt - vEyePt;
                nLook.Normalize();
                Vector3 nUp = vUp;
                nUp.Normalize();
                Vector3 nRight = nLook;
                nRight = Vector3.Cross(nRight, nUp);
                nRight.Normalize();
                nRight *= m_eyeDistance / 2.0f;
                Vector3 lookAt = nLook;

                if (m_currentEye == Eye.Left)
                {
                    vEyePt = vEyePt - nRight;
                    if (m_projectToScreen)
                    {
                        UpdateProjectionMatrix(
                            (-1) * nRight, new Vector3(0, 0, 0), displaySize);
                    }
                }
                else
                {
                    vEyePt = vEyePt + nRight;
                    if (m_projectToScreen)
                    {
                        UpdateProjectionMatrix(
                            nRight, new Vector3(0, 0, 0), displaySize);
                    }
                }
                // Calc the view matrix
                m_viewMatrix = Matrix.CreateLookAt(vEyePt, vLookatPt, vUp);
            }
        }
        /// <summary>
        /// Calculate projection matrix
        /// </summary>
        public void UpdateProjectionMatrix()
        {
           m_projectionMatrix = Matrix.CreatePerspectiveFieldOfView(
                m_fieldOfView,
                m_aspectRatio,
                m_nearPlaneDistance,
                m_farPlaneDistance);
        }
        /// <summary>
        /// Create sheared view frustum to project the
        /// camera view to a scree (tracked stereo screen projection)
        /// </summary>
        /// <param name="eyeOffset"></param>
        /// <param name="screenRotation"></param>
        /// <param name="screenPosition"></param>
        /// <param name="displaySize"></param>
        public void UpdateProjectionMatrix(
            Vector3 eyeOffset,
            Vector3 screenPosition,
            Vector2 displaySize)
        {
            
            Vector3 position = m_eyePosition;
            // TODO: screen position
            position -= eyeOffset;
            float nearPlaneWidth = (displaySize.X * m_nearPlaneDistance) / position.Z;
            float nearPlaneWidthHalf = nearPlaneWidth * 0.5f;
            float left = -nearPlaneWidthHalf;
            float right = nearPlaneWidthHalf;
            float nearPlaneHeightHalf = nearPlaneWidthHalf/m_aspectRatio;
            float down = -nearPlaneHeightHalf;
            float up = nearPlaneHeightHalf;
            
            m_projectionMatrix  = Matrix.CreatePerspectiveOffCenter(
                left, right,
                down, up,
                m_nearPlaneDistance, m_farPlaneDistance);
        }
    }
}
