/*
 * Copyright (C): Timo Fleisch
 *                2009 Berlin, Germany
 *                http://www.vrhome.de
 *                All Rights Reserved.
 *
 * THIS SOURCECODE IS PUBLISHED UNDER RESTRICTIONS.
 * No unauthorized copy or distribution allowed.
 * Please see website for detailed liscence information.
 * 
 */
using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace Tgex.Graphics
{
    public class LineInterlacedRenderWindow : RenderWindow
    {
        protected RenderTarget2D m_leftSceneRenderTarget = null;
        protected RenderTarget2D m_rightSceneRenderTarget = null;

        protected Texture2D m_leftSceneTexture;
        protected Texture2D m_rightSceneTexture;

        protected Effect m_stereoShader = null;
        protected EffectParameter m_leftSceneTexturePar = null;
        protected EffectParameter m_rightSceneTexturePar = null;
        protected EffectParameter m_heightPar = null;
        protected EffectParameter m_isAnaglyphPar = null;

        /// <summary>
        /// Vertex buffer to render stuff on screen.
        /// </summary>
        private VertexBuffer m_vbScreen;
        private VertexDeclaration m_vertexDeclaration;

        public LineInterlacedRenderWindow(VRGame game)
            : base(game)
        {
#if WINDOWS
            m_mainWin.StartPosition =
                System.Windows.Forms.FormStartPosition.Manual;
            m_mainWin.Location = new System.Drawing.Point(
                m_game.GameSettings.windowPositionX, 
                m_game.GameSettings.windowPositionY);
            m_mainWin.Height = m_height;
            m_mainWin.Width = m_width;
#endif
            // prepare graphics settings
            if (m_antiAlias) m_game.GraphicsManager.PreferMultiSampling = true;
        }

        public override void Initialize()
        {
            base.Initialize();

            // set stereo viewports
            m_defaultViewport = m_game.GraphicsDevice.Viewport;
#if WINDOWS
            m_mainWin.Location = new System.Drawing.Point(
                m_game.GameSettings.windowPositionX,
                m_game.GameSettings.windowPositionY);
            if (!m_fullscreen)
            {
                m_mainWin.ControlBox = false;
                m_mainWin.SizeGripStyle = System.Windows.Forms.SizeGripStyle.Hide;
                m_mainWin.FormBorderStyle = System.Windows.Forms.FormBorderStyle.None;
            }
#endif
            GraphicsDeviceReset(null, EventArgs.Empty);

            m_leftSceneRenderTarget = CreateRenderTarget();
            m_rightSceneRenderTarget = CreateRenderTarget();

            // load stereo shader
            // Load shader
            m_stereoShader = m_game.Content.Load<Effect>("shader/stereo");

            m_leftSceneTexturePar = m_stereoShader.Parameters["txLeft"];
            m_rightSceneTexturePar = m_stereoShader.Parameters["txRight"];
            m_heightPar = m_stereoShader.Parameters["ySize"];
            m_isAnaglyphPar = m_stereoShader.Parameters["anaglyph"];

            m_heightPar.SetValue(m_height);
            m_isAnaglyphPar.SetValue(m_game.GameSettings.anaglyph);

            // create vertices for all screen render
            VertexPositionTexture[] vertices = new VertexPositionTexture[]
                {
                    new VertexPositionTexture(
                        new Vector3(-1.0f, -1.0f, 0.5f),
                        new Vector2(0, 1)),
                    new VertexPositionTexture(
                        new Vector3(-1.0f, 1.0f, 0.5f),
                        new Vector2(0, 0)),
                    new VertexPositionTexture(
                        new Vector3(1.0f, -1.0f, 0.5f),
                        new Vector2(1, 1)),
                    new VertexPositionTexture(
                        new Vector3(1.0f, 1.0f, 0.5f),
                        new Vector2(1, 0)),
                };
            m_vbScreen = new VertexBuffer(
                m_game.GraphicsDevice,
                typeof(VertexPositionTexture),
                vertices.Length,
                BufferUsage.WriteOnly);

            m_vbScreen.SetData(vertices);

            m_vertexDeclaration = new VertexDeclaration(
                m_game.GraphicsDevice, VertexPositionTexture.VertexElements);
        }

        /// <summary>
        /// Graphics device reset handler.
        /// </summary>
        public override void GraphicsDeviceReset(object sender, EventArgs e)
        {
            // Update width and height
            m_width = m_game.GraphicsDevice.Viewport.Width;

            m_height = m_game.GraphicsDevice.Viewport.Height;
            m_game.Camera.AspectRatio = (float)m_width / (float)m_height;
            m_game.Camera.UpdateProjectionMatrix();

            // Re-Set device
            // Restore z buffer state
            m_game.GraphicsDevice.RenderState.DepthBufferEnable = true;
            m_game.GraphicsDevice.RenderState.DepthBufferWriteEnable = true;
            // Set u/v addressing back to wrap
            m_game.GraphicsDevice.SamplerStates[0].AddressU = TextureAddressMode.Wrap;
            m_game.GraphicsDevice.SamplerStates[0].AddressV = TextureAddressMode.Wrap;

            // Set 128 and greate alpha compare for Model.Render
            m_game.GraphicsDevice.RenderState.ReferenceAlpha = 128;
            m_game.GraphicsDevice.RenderState.AlphaFunction = CompareFunction.Greater;
#if WINDOWS
            if (!m_fullscreen)
            {
                m_mainWin.Height = m_height;
                m_mainWin.Width = m_width;
                m_mainWin.Location = new System.Drawing.Point(
                    m_game.GameSettings.windowPositionX,
                    m_game.GameSettings.windowPositionY);
            }
#endif
            // update shader
            if (m_heightPar != null)
            {
                //m_heightPar.SetValue(m_height);
            }
        }

        override public void UpdateStereo()
        {
            base.UpdateStereo();

            if (m_isLeft != m_switchLeftRight)
            {
                SetRenderTarget(m_leftSceneRenderTarget);
            }
            else
            {
                SetRenderTarget(m_rightSceneRenderTarget);
            }
        }
        Texture2D GetRenderTexture()
        {
            if (m_isLeft != m_switchLeftRight)
            {
                return m_leftSceneRenderTarget.GetTexture();
            }
            else
            {
                return m_rightSceneRenderTarget.GetTexture();
            }
        }

        /// <summary>
        /// Draw
        /// </summary>
        /// <param name="gameTime">Game time</param>
        public override void Draw(GameTime gameTime)
        {
            // left eye
            m_isLeft = true;
            UpdateStereo();
            m_game.DrawScene(gameTime);
            // right eye
            m_isLeft = false;
            UpdateStereo();
            m_game.DrawScene(gameTime);
            ResetRenderTarget();

            // render stereo shader
            m_leftSceneTexturePar.SetValue(m_leftSceneRenderTarget.GetTexture());
            m_rightSceneTexturePar.SetValue(m_rightSceneRenderTarget.GetTexture());
            m_stereoShader.CommitChanges();

            // Start shader
            m_stereoShader.CurrentTechnique = m_stereoShader.Techniques["Render"];
            try
            {
                m_stereoShader.Begin(SaveStateMode.None);

                for (int num = 0; num < m_stereoShader.CurrentTechnique.Passes.Count; num++)
                {
                    EffectPass pass = m_stereoShader.CurrentTechnique.Passes[num];

                    pass.Begin();
                    // render complete scene
                    m_game.GraphicsDevice.VertexDeclaration = m_vertexDeclaration;
                    m_game.GraphicsDevice.Vertices[0].SetSource(m_vbScreen, 0,
                        VertexPositionTexture.SizeInBytes);
                    m_game.GraphicsDevice.DrawPrimitives(PrimitiveType.TriangleStrip, 0, 2);
                    pass.End();
                }
            }
            finally
            {
                // End shader
                m_stereoShader.End();
            }

        }

        private RenderTarget2D CreateRenderTarget()
        {
            // Create render target of specified size.
            RenderTarget2D renderTarget = new RenderTarget2D(
                m_game.GraphicsDevice,
                m_width,
                m_height,
                1,
                SurfaceFormat.Color,
                m_game.GameSettings.multiSampleType,
                m_game.GameSettings.multiSampleQuality,
                RenderTargetUsage.PlatformContents);
/*
            zBufferSurface =
                    new DepthStencilBuffer(
                    BaseGame.Device,
                    texWidth, texHeight,
                    // Lets use the same stuff as the back buffer.
                    BaseGame.BackBufferDepthFormat,
                    // Don't use multisampling, render target does not support that.
                    MultiSampleType.None, 0);
 * */
            return renderTarget;
        }
    }
}
