﻿/*
 * Copyright (C): Timo Fleisch
 *                2009 Berlin, Germany
 *                http://www.vrhome.de
 *                All Rights Reserved.
 *
 * THIS SOURCECODE IS PUBLISHED UNDER RESTRICTIONS.
 * No unauthorized copy or distribution allowed.
 * Please see website for detailed liscence information.
 * 
 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Runtime.InteropServices;
using System.Threading;
using Emgu.CV;
using Emgu.Util;
using Emgu.CV.CvEnum;
using Microsoft.Xna.Framework;

namespace Tvrx
{

    public struct RGBQUAD
    {
        public byte rgbBlue;
        public byte rgbGreen;
        public byte rgbRed;
        public byte rgbReserved;
    }

    public struct BITMAPINFOHEADER
    {
        public int biSize;
        public int biWidth;
        public int biHeight;
        public short biPlanes;
        public short biBitCount;
        public int biCompression;
        public int biSizeImage;
        public int biXPelsPerMeter;
        public int biYPelsPerMeter;
        public int biClrUsed;
        public int biClrImportant;
    }

    public struct BITMAPINFO
    {
        public BITMAPINFOHEADER bmiHeader;
        public RGBQUAD bmiColors;
    }

    unsafe public class PGRFlyCaptureCamera : ICaptureCamera
    {
        //
        // DLL Functions to import
        // 
        // Follow this format to import any DLL with a specific function.
        //

        [DllImport("pgrflycapture.dll")]
        public static extern int flycaptureCreateContext(int* flycapcontext);

        [DllImport("pgrflycapture.dll")]
        public static extern int flycaptureStart(int flycapcontext,
            FlyCaptureVideoMode videoMode,
            FlyCaptureFrameRate frameRate);

        [DllImport("pgrflycapture.dll")]
        public static extern string flycaptureErrorToString(int error);

        [DllImport("pgrflycapture.dll")]
        public static extern int flycaptureInitialize(int flycapContext,
            int cameraIndex);

        [DllImport("pgrflycapture.dll")]
        public static extern int flycaptureGetCameraInformation(int flycapContext,
            ref FlyCaptureInfo arInfo);

        [DllImport("pgrflycapture.dll")]
        unsafe public static extern int flycaptureGrabImage2(int flycapContext,
            ref FlyCaptureImage image);

        [DllImport("pgrflycapture.dll")]
        unsafe public static extern int flycaptureSaveImage(int flycapContext,
            ref FlyCaptureImage image, string filename,
            FlyCaptureImageFileFormat fileFormat);

        [DllImport("pgrflycapture.dll")]
        public static extern int flycaptureStop(int flycapContext);

        [DllImport("pgrflycapture.dll")]
        public static extern int flycaptureDestroyContext(int flycapContext);

        [DllImport("pgrflycapture.dll")]
        public static extern int flycaptureConvertImage(int flycapContext,
            ref FlyCaptureImage image, ref FlyCaptureImage imageConvert);

        [DllImport("pgrflycapture.dll")]
        public static extern int flycaptureSetCameraProperty(
                int flycapContext,
                FlyCaptureProperty cameraProperty,
                long lValueA,
                long lValueB,
                bool bAuto);

        [DllImport("gdi32.dll")]
        public static extern int CreateDIBSection(int* hDC,
            ref BITMAPINFO pBitmapInfo, int un, ref byte* lplpVoid,
            int handle, int dw);


        // Bitmap constant
        public const short DIB_RGB_COLORS = 0;

        // The index of the camera to grab from.
        public const int _CAMERA_INDEX = 0;

        // The maximum number of cameras on the bus.
        public const int _MAX_CAMS = 3;

        // The number of images to grab.
        public const int _IMAGES_TO_GRAB = 10;


        // Function to initialize a bitmap for the converted image.
        static void InitBitmapStructure(int nRows, int nCols,
            ref FlyCaptureImage flycapRGBImage)
        {

            BITMAPINFO bmi = new BITMAPINFO();	// bitmap header
            byte* pvBits = null; //	pointer	to DIB section
            int nBytes;

            nBytes = nRows * nCols * 3;	// for R,G,B

            bmi.bmiHeader.biSize = 40; // sizeof(BITMAPINFOHEADER) = 40
            bmi.bmiHeader.biWidth = nCols;
            bmi.bmiHeader.biHeight = nRows;
            bmi.bmiHeader.biPlanes = 1;
            bmi.bmiHeader.biBitCount = 24; // three	8-bit components
            bmi.bmiHeader.biCompression = 0; //	BI_RGB = 0
            bmi.bmiHeader.biSizeImage = nBytes;
            CreateDIBSection(null, ref bmi, DIB_RGB_COLORS,
                ref pvBits, 0, 0);

            // This	is where we	set	the	"to	be converted" image	data pointer 
            // to our newly created bitmap data pointer
            flycapRGBImage.pData = pvBits;
        }

        static void reportError(int ret, string fname)
        {
            /*
            fname += " error: " + flycaptureErrorToString(ret) + "\n";
            Console.Write(fname);
            Console.Write("\nPress Enter");
            Console.Read();
            */
            return;
        }

        #region variables
        int flycapContext;
        int ret;
        FlyCaptureInfo flycapInfo = new FlyCaptureInfo();
        FlyCaptureImage image = new FlyCaptureImage();
        FlyCaptureImage flycapRGBImage = new FlyCaptureImage();
        bool doCapture = false;
        CaptureCameraDesc m_cameraDesc;
        private int m_Width;
        private int m_Height;
        #endregion

        #region properties
        /// <summary>
        /// Width of camera capture
        /// </summary>
        public int Width
        {
            get { return m_Width; }
            set { m_Width = value; }
        }
        /// <summary>
        /// Height of camera capture
        /// </summary>
        public int Height
        {
            get { return m_Height; }
            set { m_Height = value; }
        }
        #endregion

        public void SetExposure(int value)
        {
		    flycaptureSetCameraProperty( flycapContext,
									FlyCaptureProperty.FLYCAPTURE_AUTO_EXPOSURE,
									(long) value,
									0,
									false);
        }

        public void SetBrightness( int value )
        {
		    flycaptureSetCameraProperty( flycapContext, 
									     FlyCaptureProperty.FLYCAPTURE_BRIGHTNESS,
                                        (long)value,
									    0,
									    false);
        }

        public void SetGain( int value )
        {
		    flycaptureSetCameraProperty( flycapContext,
									    FlyCaptureProperty.FLYCAPTURE_GAIN,
                                        (long)value,
									    0,
									    false);
	    }

        public void SetShutter(int value)
        {
            flycaptureSetCameraProperty(flycapContext,
                                    FlyCaptureProperty.FLYCAPTURE_SHUTTER,
                                    (long)value,
                                    0,
                                    false);
        }

        //static void Main(string[] args)
        public void Initialize(CaptureCameraDesc cameraDesc)
        {
            m_cameraDesc = cameraDesc;
            int flycapContextTmp;
            // Create the context.
            ret = flycaptureCreateContext(&flycapContextTmp);
            if (ret != 0)
            {
                reportError(ret, "flycaptureCreateContext");
                return;
            }
            flycapContext = flycapContextTmp;
            // Initialize the camera.
            ret = flycaptureInitialize(flycapContext, _CAMERA_INDEX);
            if (ret != 0)
            {
                reportError(ret, "flycaptureInitialize");
                return;
            }

            // Get the info for this camera.
            ret = flycaptureGetCameraInformation(flycapContext, ref flycapInfo);
            if (ret != 0)
            {
                reportError(ret, "flycaptureGetCameraInformation");
                return;
            }
            if (flycapInfo.CameraType ==
                FlyCaptureCameraType.FLYCAPTURE_BLACK_AND_WHITE)
            {
                Console.Write("Model: B&W " + flycapInfo.pszModelString + "\n"
                    + "Serial #: " + flycapInfo.SerialNumber + "\n");
            }
            else if (flycapInfo.CameraType ==
                FlyCaptureCameraType.FLYCAPTURE_COLOR)
            {
                Console.Write("Model: Colour " + flycapInfo.pszModelString + "\n"
                    + "Serial #: " + flycapInfo.SerialNumber + "\n");
            }
            InitBitmapStructure(cameraDesc.width, cameraDesc.height, ref flycapRGBImage);
            flycapRGBImage.pixelFormat = FlyCapturePixelFormat.FLYCAPTURE_BGR;

        }
        /*
        public Byte[] GetImage()
        {
            lock (imageLock)
            {
                int imgSize = flycapRGBImage.iCols * flycapRGBImage.iRows * 3;
                Marshal.Copy((IntPtr)flycapRGBImage.pData, imgCopy, 0, imgSize);
            }
            return imgCopy;
        }
        */

        public void StartCapture()
        {
            // Start FlyCapture
            FlyCaptureVideoMode videoMode
                = FlyCaptureVideoMode.FLYCAPTURE_VIDEOMODE_640x480Y8;
            if (m_cameraDesc.colorMode == ColorMode.Y8)
            {
                if (m_cameraDesc.width == 640)
                {
                    videoMode = FlyCaptureVideoMode.FLYCAPTURE_VIDEOMODE_640x480Y8;
                }
                else if (m_cameraDesc.width == 800)
                {
                    videoMode = FlyCaptureVideoMode.FLYCAPTURE_VIDEOMODE_800x600Y8;
                }
                else if (m_cameraDesc.width == 1024)
                {
                    videoMode = FlyCaptureVideoMode.FLYCAPTURE_VIDEOMODE_1024x768Y8;
                }
            }
            ret = flycaptureStart(flycapContext,
                videoMode,
                FlyCaptureFrameRate.FLYCAPTURE_FRAMERATE_30);
            if (ret != 0)
            {
                reportError(ret, "flycaptureStart");
                return;
            }
            doCapture = true;
        }

        public void SetProperty(FlyCaptureProperty prop, int value)
        {
            // set property
            flycaptureSetCameraProperty(flycapContext,
                                    prop,
                                    (long)value,
                                    0,
                                    false);
        }

        unsafe public IntPtr GrabImage()
        {
            if (doCapture)
            {
                ret = flycaptureGrabImage2(flycapContext, ref image);
                if (ret != 0)
                {
                    reportError(ret, "flycaptureGrabImage2");
                    return System.IntPtr.Zero;
                }
                return (System.IntPtr)image.pData;
            }
            return System.IntPtr.Zero;
        }
        /*
        unsafe public void GrabImage()
        {
            MCvPoint2D32f[] imagePointsTemp = new MCvPoint2D32f[255];
            IntPtr storage = new IntPtr();
            storage = CvInvoke.cvCreateMemStorage(1000);
            while (doCapture)
            {
                ret = flycaptureGrabImage2(flycapContext, ref image);
                if (ret != 0)
                {
                    reportError(ret, "flycaptureGrabImage2");
                    return;
                }
                MIplImage source = new MIplImage();
                //MIplImage dest = new MIplImage();
                IntPtr sourcePtr = new IntPtr();
                //IntPtr destPtr = new IntPtr();

                //source.depth = Emgu.CV.CvEnum.IPL_DEPTH.IPL_DEPTH_8U;
                //source.height = 480;
                //source.imageSize = 480 * 640;
                //source.width = 640;
                //source.align = 4;
                //source.imageData = (System.IntPtr) image.pData;

                //Image<Gray, Byte> sourceImage = new Image<Gray, Byte>(source.width, source.height);
                //Image<Bgr, Byte> destImage = new Image<Bgr, Byte>(640,480);
                MCvSize size = new MCvSize(640,480);
                sourcePtr = CvInvoke.cvCreateImage(size, Emgu.CV.CvEnum.IPL_DEPTH.IPL_DEPTH_8U, 1);
                source = (MIplImage)Marshal.PtrToStructure(sourcePtr, typeof(MIplImage));
                source.imageData = (System.IntPtr)image.pData;
                Marshal.StructureToPtr(source, sourcePtr, false);

                //CvInvoke.cvCvtColor(sourcePtr, destImage.Ptr, Emgu.CV.CvEnum.COLOR_CONVERSION.CV_GRAY2BGR);
                //dest = (MIplImage)Marshal.PtrToStructure(destImage.Ptr, typeof(MIplImage));

                // find circle
                int numCircles = 0;
                double globalThreshold = 100.0;

                int header_size, i, count;
                MCvSeq contour = new MCvSeq();

                header_size = 88; // hopefully
                MCvSize iiSize = new MCvSize();
                iiSize.height = source.height;
                iiSize.width = source.width;
                IntPtr m_tmp8uC1 = CvInvoke.cvCreateImage(iiSize, IPL_DEPTH.IPL_DEPTH_8U, 1);

                CvInvoke.cvThreshold(sourcePtr, m_tmp8uC1, globalThreshold, 255, THRESH.CV_THRESH_BINARY);
                IntPtr contourPtr = new IntPtr(); ;
                //Marshal.StructureToPtr(*contour,contourPtr, false);
                CvInvoke.cvFindContours(m_tmp8uC1, storage, ref contourPtr,
                                  header_size, RETR_TYPE.CV_RETR_EXTERNAL,
                                  CHAIN_APPROX_METHOD.CV_CHAIN_APPROX_SIMPLE,new MCvPoint(0,0));
                // GO THROUGH ALL CONTOURS
                while ((contourPtr != IntPtr.Zero) && (numCircles < 255))
                {
                    contour = (MCvSeq)Marshal.PtrToStructure(contourPtr, typeof(MCvSeq));
                    int CV_SEQ_KIND_MASK = (((1 << 3) - 1) << 9);
                    int CV_SEQ_KIND_CURVE = (1 << 9);
                    if (((contour).flags & CV_SEQ_KIND_MASK) == CV_SEQ_KIND_CURVE)
                    {
                        // HOW MANY POINTS HAS THE CONTOUR ??
                        count = contour.total;

                        //if (count >=6)
                        if ((count >= 6) && (count <= 35))
                        {
                            MCvBox2D circleBox = CvInvoke.cvFitEllipse2(contourPtr);
                            imagePointsTemp[numCircles].x = (float)circleBox.center.x;
                            imagePointsTemp[numCircles].y = (float)circleBox.center.y;
                            numCircles++;
                        }

                        else if ((count <= 35))
                        {
                            MCvRect rect = CvInvoke.cvBoundingRect(contourPtr,true);
                            imagePointsTemp[numCircles].x = (float)rect.x + rect.width / 2.0f;
                            imagePointsTemp[numCircles].y = (float)rect.y + rect.height / 2.0f;
                            numCircles++;
                        }

                    }
                    // GOT TO THE NEXT CONTOUR (IF ANY)
                    contourPtr = contour.h_next;
                }

                lock (imagePointsLock)
                {
                    m_numOfPoints = numCircles;
                    for (i = 0; i < numCircles; i++)
                    {
                        imagePoints[i].X = imagePointsTemp[i].x;
                        imagePoints[i].Y = imagePointsTemp[i].y;
                    }
                }
                //    if (contour!=NULL) free (contour);
                CvInvoke.cvReleaseImage(ref m_tmp8uC1);
                CvInvoke.cvReleaseImage(ref sourcePtr);

                //--------------------------------------------------
                if (ret != 0)
                {
                    reportError(ret, "flycaptureConvertImage");
                    return;
                }
            }
            CvInvoke.cvReleaseMemStorage(ref storage);
        }
*/
        /*

            // Save the image.
            Console.Write("\nSaving Last Image ");
            ret = flycaptureSaveImage(flycapContext, ref flycapRGBImage, "raw.bmp",
                FlyCaptureImageFileFormat.FLYCAPTURE_FILEFORMAT_BMP);
            if (ret != 0)
            {
                reportError(ret, "flycaptureSaveImage");
                return;
            }
            else
                System.Diagnostics.Process.Start("mspaint.exe", "raw.bmp");
        */
        public void StopCapture()
        {
            doCapture = false;
            //captureThread.Abort();
            // Stop FlyCapture.
            ret = flycaptureStop(flycapContext);
            if (ret != 0)
            {
               // reportError(ret, "flycaptureStop");
                return;
            }
        }

        public void CloseCamera()
        {
            // Destroy the context.
            ret = flycaptureDestroyContext(flycapContext);
            if (ret != 0)
            {
                reportError(ret, "flycaptureDestroyContext");
                return;
            }
        }
    }
}
