﻿/*
 * Copyright (C): Timo Fleisch
 *                2009 Berlin, Germany
 *                http://www.vrhome.de
 *                All Rights Reserved.
 *
 * THIS SOURCECODE IS PUBLISHED UNDER RESTRICTIONS.
 * No unauthorized copy or distribution allowed.
 * Please see website for detailed liscence information.
 * 
 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using Microsoft.Xna.Framework;
using Emgu.CV;
using Emgu.Util;
using Emgu.CV.CvEnum;

namespace Tvrx
{
    public class Posit : IPoseEstimate
    {
        IntPtr m_positObject;
        MCvPoint2D32f[] m_imagePoints;
        Object m_imagePointsLock = new Object();
        Vector2[] m_imagePointsBuffer;
        int m_numOfMarker;
        bool m_imagePointsUpdate = false;
        bool m_doPosit = false;
        Thread m_positThread;
        double m_focalLengthMM;
        bool m_flipImage;
        float m_scale;

        // to swap camera axis
        int[] m_assignAxis = new int[3];
        int[] m_assignAxisSign = new int[3];

        Object m_resultLock = new Object();
        Vector3 m_positionBuffer;
        Vector3 m_rotationBuffer;

        /// <summary>
        /// If new marker position is far away, increase error count. Only after certain amount
        /// of counts accept new position.
        /// </summary>
        protected int m_errorCount;
        /// <summary>
        /// Calculated position from last capture frame.
        /// </summary>
        protected Vector3 m_previousPosition;
        /// <summary>
        /// Calculated roation from last capture frame.
        /// </summary>
        protected Vector3 m_previousRotation;
        /// <summary>
        /// Angles are circulare. But to be able to make filtering between 0 and 2pi,
        /// we continue to count up or down. Therefore for each calculated yaw angle
        /// we add this offset.
        /// </summary>
        protected float m_YawOffset;

        unsafe ~Posit()
        {
            IntPtr p = m_positObject;
            CvInvoke.cvReleasePOSITObject((IntPtr)(&p));
        }

        public void InitializeCameraParameter(double focalLengthMM, bool flipImage, float scale,
            int[] assingAxis, int[] assignAxisSign)
        {
            m_focalLengthMM = focalLengthMM;
            m_flipImage = flipImage;
            m_scale = scale;
            m_assignAxis = assingAxis;
            m_assignAxisSign = assignAxisSign;
        }
        public void InitializeMarkerBody(Vector3[] markerPoints)
        {
            m_numOfMarker = markerPoints.Length;
            MCvPoint3D32f[] worldMarker = new MCvPoint3D32f[m_numOfMarker];
            for (int i = 0; i < m_numOfMarker; i++)
            {
                worldMarker[i].x = markerPoints[i].X;
                worldMarker[i].y = markerPoints[i].Y;
                worldMarker[i].z = markerPoints[i].Z;
            }

            m_positObject = CvInvoke.cvCreatePOSITObject(
                worldMarker, m_numOfMarker);

            m_imagePoints = new MCvPoint2D32f[m_numOfMarker];
            m_imagePointsBuffer = new Vector2[m_numOfMarker];

        }
        public void UpdateImagePoints(Vector2[] imagePoints)
        {
            Monitor.Enter(m_imagePointsLock);
            for (int i = 0; i < m_numOfMarker; i++)
            {
                m_imagePointsBuffer[i] = imagePoints[i];
            }
            m_imagePointsUpdate = true;
            Monitor.Pulse(m_imagePointsLock);
            Monitor.Exit(m_imagePointsLock);
        }

        public void GetTransform(out Vector3 position, out Vector3 rotation)
        {
            lock (m_resultLock)
            {
                position = m_positionBuffer;
                rotation = m_rotationBuffer;
            }
        }
        public void StartEstimation()
        {
            m_doPosit = true;
            m_positThread = new Thread(PoseEstimate);
            m_positThread.Start();
        }
        public void StopEstimation()
        {
            m_doPosit = false;
            // to avoid stop because of wait state
            if (Monitor.TryEnter(m_imagePointsLock))
            {
                Monitor.Pulse(m_imagePointsLock);
                Monitor.Exit(m_imagePointsLock);
            }
        }

        #region POSIT
        float[] POSITRot = new float[9];
        float[] POSITTrans = new float[3];
        float[] swapAxis = new float[3];
        double[,] POSITRot1 = new double[3, 3];
        float[] angle = new float[3];

        unsafe private void PoseEstimate()
        {
            while (m_doPosit)
            {
                // copy image points
                Monitor.Enter(m_imagePointsLock);
                if (!m_imagePointsUpdate)
                {
                    Monitor.Wait(m_imagePointsLock);
                }
                for (int i = 0; i < m_numOfMarker; i++)
                {
                    m_imagePoints[i].x = m_imagePointsBuffer[i].X;
                    m_imagePoints[i].y = m_imagePointsBuffer[i].Y;
                }
                m_imagePointsUpdate = false;
                Monitor.Exit(m_imagePointsLock);

                MCvTermCriteria criteria;
                criteria.type = TERMCRIT.CV_TERMCRIT_EPS | TERMCRIT.CV_TERMCRIT_ITER;
                criteria.epsilon = 0.00001;
                criteria.max_iter = 500;
                CvInvoke.cvPOSIT(
                    m_positObject,
                    m_imagePoints, m_focalLengthMM,
                    criteria, POSITRot, POSITTrans);


                if (m_flipImage)
                {
                    // for unkown reason posit algorithm negates y-axis, therefore negate again
                    // perhaps definition of marker confuses y-axis
                    // if image has been fliped arround y (should be the correct way)
                    // then this has to be done
                    POSITTrans[1] *= -1;
                }

                // swap axis according to config file spec
                
                //float[] swapAxis = new float[3];
                // swap translation
                for (int i=0; i<3;i++) swapAxis[i] = POSITTrans[i];
                for (int i=0; i<3;i++) POSITTrans[i] = (-1) * m_assignAxisSign[i]*swapAxis[m_assignAxis[i]];

                // first check if the new value is not to different - we assume error then
                float error = SquareError(POSITTrans, m_previousPosition);
                if ((error > 5000) && (m_errorCount < 10))
                {
                    m_errorCount++;
                }
                else
                {
                    // *******************************************************
                    // Translation
                    // *******************************************************
                    Vector3 position;
                    position.X = POSITTrans[0];
                    position.Y = POSITTrans[1];
                    position.Z = POSITTrans[2];
                    position *= m_scale;
                    // *******************************************************
                    // Rotation
                    // *******************************************************

                    //double[,] POSITRot1 = new double[3, 3];
                    for (int i = 0; i < 3; i++)
                        for (int j = 0; j < 3; j++)
                            POSITRot1[i, j] = POSITRot[i * 3 + j];
                    // euler angles
                    float phi1 = 0;
                    float tchi1 = 0;
                    float teta1 = 0;
                    // teta = Z, phi = X, Y = -tchi
                    EulerAngles(POSITRot1, ref teta1, ref phi1, ref tchi1);

                    // phi is rotated by pi already - inverse
                    phi1 -= (float)Math.PI;

                    // calculate angle difference
                    float diffTchi = tchi1 - m_previousRotation.Y;
                    m_previousRotation = new Vector3(phi1, tchi1, teta1);
                    // Angles are circulare. But to be able to make filtering between 0 and 2pi,
                    // we continue to count up or down. Therefore for each calculated yaw angle
                    // we add this offset.
                    if (diffTchi > 4.712389)
                        m_YawOffset -= (float)(2 * Math.PI);
                    if (diffTchi < -4.712389)
                        m_YawOffset += (float)(2 * Math.PI);
                    tchi1 += m_YawOffset;

                    // stop confusion - give angle names according to camera coordinate system
                    //float[] angle = new float[3];
                    angle[0] = phi1;
                    angle[1] = teta1;
                    angle[2] = tchi1;

                    // swap rotaiton axis according to config file definition - same as with translation before
                    for (int i=0; i<3;i++) swapAxis[i] = angle[i];
                    for (int i=0; i<3;i++) angle[i] = m_assignAxisSign[i]*swapAxis[m_assignAxis[i]];
                    Vector3 rotation;
                    rotation.X = angle[0];
                    rotation.Y = angle[1];
                    rotation.Z = angle[2];
                    if (m_flipImage)
                    {
                        // for translation we had to negate y - for rotation we have to negate x,z
                        // reason unkown...
                        // use this for fliped image (arround y)
                        rotation.X *= -1;
                        rotation.Z *= -1;
                    }
                    // write to buffer
                    lock (m_resultLock)
                    {
                        m_positionBuffer = position;
                        m_rotationBuffer = rotation;
                    }
                }
            }
        }
        private float SquareError(float[] a, Vector3 b)
        {
            double result = 0;
            result += Math.Pow(a[0] - b.X, 2);
            result += Math.Pow(a[1] - b.Y, 2);
            result += Math.Pow(a[2] - b.Z, 2);
            return (float)result;
        }
        // according to: http://www.kwon3d.com/theory/euler/euler_angles.html
        // http://www.euclideanspace.com/maths/geometry/rotations/conversions/matrixToEuler/index.htm
        // teta = Z, phi = X, Y = -tchi
        //void eulerAngles( double m[3][3], float &heading, float &bank, float &attitude ) 
        private void EulerAngles(double[,] POSITRot1, ref float teta, ref float phi, ref float tchi)
        {

            teta = (float)Math.Asin(POSITRot1[2, 0]);
            if (POSITRot1[2, 2] * Math.Cos(teta) > 0)
            {
                phi = (float)Math.Atan(-POSITRot1[2, 1] / POSITRot1[2, 2]);
            }
            else
            {
                phi = (float)(Math.Atan(-POSITRot1[2, 1] / POSITRot1[2, 2]) + Math.PI);
            }
            if (POSITRot1[0, 0] * Math.Cos(teta) > 0)
            {
                tchi = (float)Math.Atan(-POSITRot1[1, 0] / POSITRot1[0, 0]);
            }
            else
            {
                tchi = (float)(Math.Atan(-POSITRot1[1, 0] / POSITRot1[0, 0]) + Math.PI);
            }
        }
        #endregion

    }
}
