﻿/*
 * Copyright (C): Timo Fleisch
 *                2009 Berlin, Germany
 *                http://www.vrhome.de
 *                All Rights Reserved.
 *
 * THIS SOURCECODE IS PUBLISHED UNDER RESTRICTIONS.
 * No unauthorized copy or distribution allowed.
 * Please see website for detailed liscence information.
 * 
 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Runtime.InteropServices;
using Emgu.CV;
using Emgu.Util;
using Emgu.CV.CvEnum;
using Microsoft.Xna.Framework;

namespace Tvrx
{
    public class TCamTracker : IMarkerTracker
    {
        ICaptureCamera m_camera = null;
        List<CaptureCameraDesc> m_cameraDescList;
        CaptureCameraDesc m_cameraDesc;
        Dictionary<int, IPoseEstimate> m_poseEstimateDict = new Dictionary<int,IPoseEstimate>();
        MarkerTrackerDesc m_tcamDesc;
        bool m_isTracking = false;

        // to swap camera axis
        int[] assignAxis = new int[3];
        int[] assignAxisSign = new int[3];

        Thread m_trackingThread;

        Object imageLock = new Object();
        Object m_imagePointsLock = new Object();
        Vector2[] imagePoints = new Vector2[255];

        public int m_numOfPoints { get; private set; }
        /// <summary>
        /// key is id of recognized marker and value is array of image points in right order
        /// </summary>
        Dictionary<int, Vector2[]> m_recognizedImagePoints = new Dictionary<int, Vector2[]>();
        Dictionary<int, Vector2[]> m_markerImagePointsBuffer = new Dictionary<int, Vector2[]>();

        protected List<TCamMarkerBodyDesc> m_MarkerBodyList = new List<TCamMarkerBodyDesc>();
        /// <summary>
        /// List of marker devices
        /// </summary>
        public List<TCamMarkerBodyDesc> MarkerDeviceList
        {
            get { return m_MarkerBodyList; }
            set { m_MarkerBodyList = value; }
        }

        public void RegisterMarkerBody(MarkerBodyDesc markerDesc, IPoseEstimate poseEstimate)
        {
            if (markerDesc is TCamMarkerBodyDesc)
            {
                m_MarkerBodyList.Add((TCamMarkerBodyDesc)markerDesc);
                m_markerImagePointsBuffer.Add(markerDesc.Id, new Vector2[5]);
                double focalLengthMM = m_cameraDesc.focalLengthX * m_cameraDesc.pixelSize;
                poseEstimate.InitializeCameraParameter(
                    focalLengthMM, m_cameraDesc.flipImage, m_tcamDesc.Scale,
                    assignAxis, assignAxisSign);
                m_poseEstimateDict.Add(markerDesc.Id, poseEstimate);
            }
        }

        public void getImagePoints(int id, ref Vector2[] points)
        {
            lock (m_imagePointsLock)
            {
                for (int i = 0; i < 5; i++)
                {
                    points[i] = m_markerImagePointsBuffer[id][i];
                }
            }
        }

        public TCamTracker()
        {
        }
        ~TCamTracker()
        {
            m_camera.CloseCamera();
        }

        private int Axis2Number(String s)
        {
            s = s.ToUpper();
            switch (s)
            {
                case "X":
                    return 0;
                case "Y":
                    return 1;
                case "Z":
                    return 2;
            }
            return 0;
        }
        public void Initialize(MarkerTrackerDesc desc)
        {
            m_tcamDesc = desc;
            // assign axis
			// axis sign position - if there is a minus sign, 
			// we have to read the next char as axis
			int assp = 0;
			if (desc.XAxis.Length > 0)
			{
				if (desc.XAxis.Substring(0,1) == "-")
				{
					assignAxisSign[0] = -1;
					assp++;
				}
                else
                {
                    assignAxisSign[0] = 1;
                }
				assignAxis[0] = Axis2Number(desc.XAxis.Substring(assp,1));
			}
			assp = 0;
			if (desc.YAxis.Length > 0)
			{
				if (desc.YAxis.Substring(0,1) == "-")
				{
					assignAxisSign[1] = -1;
					assp++;
				}
                else
                {
                    assignAxisSign[1] = 1;
                }
                assignAxis[1] = Axis2Number(desc.YAxis.Substring(assp,1));
			}
			assp = 0;
			if (desc.ZAxis.Length > 0)
			{
                if (desc.ZAxis.Substring(0,1) == "-")
				{
					assignAxisSign[2] = -1;
					assp++;
				}
                else
                {
                    assignAxisSign[2] = 1;
                }
                assignAxis[2] = Axis2Number(desc.ZAxis.Substring(assp,1));
			}

            Initialize();
        }
        public void Initialize()
        {
            m_cameraDescList = ConfigReader.ReadCameraDesc();
            String activeCameraId = "dragonfly 1024x768";

            foreach (CaptureCameraDesc desc in m_cameraDescList)
            {
                if (desc.id == activeCameraId)
                {
                    m_cameraDesc = desc;
                    break;
                }
            }
            if ((m_cameraDesc.model == "DragonFly")
                || (m_cameraDesc.model == "FireFlyMV"))
            {
                m_camera = new PGRFlyCaptureCamera();
            }
            else
            {
                // error out
            }

            Byte[] imgCopy = new Byte[m_cameraDesc.width * m_cameraDesc.height * 4];

            m_camera.Initialize(m_cameraDesc);

            m_camera.SetExposure(472);
            m_camera.SetBrightness(1);
            m_camera.SetGain(50);
            m_camera.SetShutter(150);

            // for testing
            /*
            List<MarkerBodyDesc> markerList = ConfigReader.ReadMarkerBodyDesc();
            foreach (MarkerBodyDesc marker in markerList)
            {
                RegisterMarkerBody(marker);
            }
             * */
        }

        public void StartTracking()
        {
            if (!m_isTracking)
            {
                m_isTracking = true;
                m_camera.StartCapture();
                m_trackingThread = new Thread(TrackingLoop);
                m_trackingThread.Start();
            }
        }
        public void StopTracking()
        {
            if (m_isTracking)
            {
                m_isTracking = false;
                // waht untill thread ends
                while (m_trackingThread.IsAlive)
                { }
                m_camera.StopCapture();
            }
        }

        unsafe public void TrackingLoop()
        {
            MCvPoint2D32f[] imagePointsTemp = new MCvPoint2D32f[255];
            IntPtr storage = new IntPtr();
            storage = CvInvoke.cvCreateMemStorage(1000);
            while (m_isTracking)
            {
                MIplImage source = new MIplImage();
                IntPtr sourcePtr = new IntPtr();

                MCvSize size = new MCvSize(m_cameraDesc.width, m_cameraDesc.height);
                sourcePtr = CvInvoke.cvCreateImage(size, Emgu.CV.CvEnum.IPL_DEPTH.IPL_DEPTH_8U, 1);
                source = (MIplImage)Marshal.PtrToStructure(sourcePtr, typeof(MIplImage));
                source.imageData = m_camera.GrabImage();
                Marshal.StructureToPtr(source, sourcePtr, false);

                // find circle
                int numCircles = 0;
                double globalThreshold = 100.0;

                int header_size, i, count;
                MCvSeq contour = new MCvSeq();

                header_size = 88; // hopefully
                MCvSize iiSize = new MCvSize();
                iiSize.height = source.height;
                iiSize.width = source.width;
                IntPtr m_tmp8uC1 = CvInvoke.cvCreateImage(iiSize, IPL_DEPTH.IPL_DEPTH_8U, 1);

                CvInvoke.cvThreshold(sourcePtr, m_tmp8uC1, globalThreshold, 255, THRESH.CV_THRESH_BINARY);
                IntPtr contourPtr = new IntPtr(); ;
                //Marshal.StructureToPtr(*contour,contourPtr, false);
                CvInvoke.cvFindContours(m_tmp8uC1, storage, ref contourPtr,
                                  header_size, RETR_TYPE.CV_RETR_EXTERNAL,
                                  CHAIN_APPROX_METHOD.CV_CHAIN_APPROX_SIMPLE, new MCvPoint(0, 0));
                // GO THROUGH ALL CONTOURS
                while ((contourPtr != IntPtr.Zero) && (numCircles < 255))
                {
                    contour = (MCvSeq)Marshal.PtrToStructure(contourPtr, typeof(MCvSeq));
                    int CV_SEQ_KIND_MASK = (((1 << 3) - 1) << 9);
                    int CV_SEQ_KIND_CURVE = (1 << 9);
                    if (((contour).flags & CV_SEQ_KIND_MASK) == CV_SEQ_KIND_CURVE)
                    {
                        // HOW MANY POINTS HAS THE CONTOUR ??
                        count = contour.total;

                        //if (count >=6)
                        if ((count >= 6) && (count <= 35))
                        {
                            MCvBox2D circleBox = CvInvoke.cvFitEllipse2(contourPtr);
                            imagePointsTemp[numCircles].x = (float)circleBox.center.x;
                            imagePointsTemp[numCircles].y = (float)circleBox.center.y;
                            numCircles++;
                        }

                        else if ((count <= 35))
                        {
                            MCvRect rect = CvInvoke.cvBoundingRect(contourPtr, true);
                            imagePointsTemp[numCircles].x = (float)rect.x + rect.width / 2.0f;
                            imagePointsTemp[numCircles].y = (float)rect.y + rect.height / 2.0f;
                            numCircles++;
                        }

                    }
                    // GOT TO THE NEXT CONTOUR (IF ANY)
                    contourPtr = contour.h_next;
                }
                CvInvoke.cvReleaseImage(ref m_tmp8uC1);
                CvInvoke.cvReleaseImage(ref sourcePtr);

                UndistortImagePoints(ref imagePointsTemp, numCircles);
                m_numOfPoints = numCircles;
                for (i = 0; i < numCircles; i++)
                {
                    imagePoints[i].X = imagePointsTemp[i].x;
                    imagePoints[i].Y = imagePointsTemp[i].y;
                }

                // finding marker bodies
                DetectBodyElements();
                //m_recognizedElements.Clear();
                m_recognizedImagePoints.Clear();
                for (int j=0; j<m_recognizedElements.Count; j++)
                {
                    int barId = m_recognizedElements[0];
                    m_recognizedElements.RemoveAt(0);
                    FindBody(barId);
                }
                // here write results to a thread save buffer
                //lock (m_imagePointsLock)
                //{
                    foreach (KeyValuePair<int, Vector2[]> kvp in m_recognizedImagePoints)
                    {
                        for (int k = 0; k < 5; k++)
                        {
                            //m_markerImagePointsBuffer[kvp.Key][k] = kvp.Value[k];
                            // send image points to pose estimator
                            m_poseEstimateDict[kvp.Key].UpdateImagePoints(kvp.Value);
                        }
                    }
                //}
                m_recognizedImagePoints.Clear();
            }
            CvInvoke.cvReleaseMemStorage(ref storage);
        }
        /// <summary>
        /// Undistorts the points according to distortion coefficients
        /// and converts from pixel to mm.
        /// </summary>
        /// <param name="imageMarkers"></param>
        /// <param name="numOfPoints"></param>
        private void UndistortImagePoints(ref MCvPoint2D32f[] imageMarkers, int numOfPoints)
        {
	        int i;
            for (i = 0; i < numOfPoints; i++)
	        {
                MCvPoint2D32f imageMarker = imageMarkers[i];
                float u0 = m_cameraDesc.principalPointX;
                float v0 = m_cameraDesc.principalPointY;
                float fx = m_cameraDesc.focalLengthX;
                float fy = m_cameraDesc.focalLengthY;
                float _fx = 1.0f / fx, _fy = 1.0f / fy;
                float k1 = -m_cameraDesc.k1, k2 = -m_cameraDesc.k2;
                float p1 = -m_cameraDesc.p1, p2 = -m_cameraDesc.p2;

                float y = (imageMarker.y - v0) * _fy;
                float y2 = y * y;
                float ky = 1 + (k1 + k2 * y2) * y2;
                float k2y = 2 * k2 * y2;
                float _2p1y = 2 * p1 * y;
                float _3p1y2 = 3 * p1 * y2;
                float p2y2 = p2 * y2;

                float x = (imageMarker.x - u0) * _fx;
                float x2 = x * x;
                float kx = (k1 + k2 * x2) * x2;
                float d = kx + ky + k2y * x2;
                float _u = fx * (x * (d + _2p1y) + p2y2 + (3 * p2) * x2) + u0;
                float _v = fy * (y * (d + (2 * p2) * x) + _3p1y2 + p1 * x2) + v0;
                imageMarker.x = _u;
                imageMarker.y = _v;

                // transform to image mm
                imageMarker.x = imageMarker.x - m_cameraDesc.principalPointX;
                imageMarker.y = imageMarker.y - m_cameraDesc.principalPointY;
                imageMarker.x *= m_cameraDesc.pixelSize;
                imageMarker.y *= m_cameraDesc.pixelSize;
                imageMarkers[i] = imageMarker;
            }
        }

        #region Find 3 marker points line

        List<Vector2> m_remainingPoints = new List<Vector2>();
        List<Vector2> m_markerVector = new List<Vector2>();
        List<int> m_recognizedElements = new List<int>();
        float m_maxLineDistance = 0.03f;
        public struct MarkerVector {
	        public Vector2 marker;
	        public float distance;
        };

        private void DetectBodyElements()
        {
	        int numOfMarkers = m_numOfPoints;
            Vector2[] imageMarker = imagePoints;

	        // reset
            foreach (TCamMarkerBodyDesc element in m_MarkerBodyList)
            {
                element.BarRecognized = false;
            }

	        m_remainingPoints.Clear();

	        // find the position of the two led bar
	        if (numOfMarkers >= 3)
	        {
		        m_markerVector.Clear();
		        //m_recognizedElements.clear();
		        for (int i=0; i<numOfMarkers; i++)
		        {
			        //Vector2 temp = imageMarker[i];
			        m_markerVector.Add( imageMarker[i] );
		        }
		        while (m_markerVector.Count >= 3)
		        {
			        // create marker vector
			        int referenceMarker = 0;
			        MarkerVector mv;
			        List<MarkerVector> tVector = new List<MarkerVector>();

			        tVector.Clear();
			        mv.marker = m_markerVector[referenceMarker];
			        mv.distance = 0;
			        tVector.Add( mv );
			        for (int v = referenceMarker+1; v<(int)m_markerVector.Count; v++)
			        {
				        mv.marker = m_markerVector[v];
				        mv.distance = ((tVector[0].marker-mv.marker)).Length();

				        int it = 0;
				        while ((tVector[it].distance < mv.distance) )
				        {
					        it++;
					        if (it==tVector.Count) break;
				        }
                        if (it == tVector.Count)
				        {
					        tVector.Add( mv );
				        }
				        else
				        {
					        tVector.Insert( it, mv );
				        }
			        }
			        bool pResult = ProcessVector ( ref tVector );
			        // if there was no led bar found, then the reference point has to be deleted
			        if (!pResult) 
			        {

				        m_remainingPoints.Add( m_markerVector[0] );
                        m_markerVector.RemoveAt(0);
			        }

		        }
		        while (m_markerVector.Count > 0)
		        {
			        m_remainingPoints.Add( m_markerVector[0] );
			        m_markerVector.RemoveAt(0);				
		        }

	        }
        }

        public bool ProcessVector( ref List<MarkerVector> tVector )
        {
	        // find permutation that is led bar (3 markers in a line)
	        for (int point1 = 1; point1 < (int)tVector.Count-1; point1++)
	        {
		        for (int point2 = point1+1; point2 < (int)tVector.Count; point2++)
		        {
			        // test reference, point1, point2
			        bool tResult = TestMarkers( tVector[0].marker, 
										        tVector[point1].marker, 
										        tVector[point2].marker );
			        if (tResult) return true;
		        }
	        }
	        return false;
        }

        // test of the three markers form a line and represent an led bar
        public bool TestMarkers( Vector2 a, Vector2 b, Vector2 c )
        {
	        // calculate the distance 
	        double dist01, dist02, dist12;

	        int startLinePoint = 0;
	        int endLinePoint = 0;
	        int onLinePoint = 0;

	        Vector2[] imageMarker = new Vector2[3];

	        imageMarker[0] = a;
	        imageMarker[1] = b;
	        imageMarker[2] = c;

	        dist01 = (imageMarker[1]-imageMarker[0]).Length();
	        dist02 = (imageMarker[2]-imageMarker[0]).Length();
	        dist12 = (imageMarker[2]-imageMarker[1]).Length();

	        // find start and end point of line
	        if ((dist01 >= dist02) && (dist01 >= dist12))
	        {
		        startLinePoint = 0;
		        endLinePoint = 1;
		        onLinePoint = 2;
	        }
	        else if ((dist02 >= dist01) && (dist02 >= dist12))
	        {
		        startLinePoint = 0;
		        endLinePoint = 2;
		        onLinePoint = 1;
	        }
	        else if ((dist12 >= dist01) && (dist12>= dist02))
	        {
		        startLinePoint = 1;
		        endLinePoint = 2;
		        onLinePoint = 0;
	        }

	        // test if onLinePoint is on the line defined by start and end point
	        // do that by calculating distance
	        // according to: http://mathenexus.zum.de/html/geometrie/abstaende/AbstandPG.htm
	        Vector2 rv = imageMarker[endLinePoint]-imageMarker[startLinePoint];

	        Vector2 p_ap = imageMarker[onLinePoint]-imageMarker[startLinePoint];
	        float dot_rv = Vector2.Dot(rv, rv);
	        float lamda = Vector2.Dot(p_ap, (rv/dot_rv));

	        Vector2 distVec = imageMarker[onLinePoint]-(imageMarker[startLinePoint] +
		        lamda * rv);

	        float distVecLength = distVec.Length();

	        // finally the point was on the line
	        if (distVecLength <= m_maxLineDistance)
	        {
		        // set the marker closer to the onLinePoint as startPoint
		        if ((imageMarker[onLinePoint]-imageMarker[startLinePoint]).Length() >
			        (imageMarker[onLinePoint]-imageMarker[endLinePoint]).Length())
		        {
			        // exchange
			        int temp = startLinePoint;
			        startLinePoint = endLinePoint;
			        endLinePoint = temp;
		        }

		        // calculate max led bar length
		        float islen = (imageMarker[endLinePoint]-imageMarker[startLinePoint]).Length();

		        // calculate led bar id
		        float markerCode = (imageMarker[onLinePoint]-imageMarker[startLinePoint]).Length()
		        / (imageMarker[endLinePoint]-imageMarker[startLinePoint]).Length();

		        float tolerance = 0.03f;

		        for (int lbId = 0; lbId < m_MarkerBodyList.Count; lbId++)
		        {
			        TCamMarkerBodyDesc element = (TCamMarkerBodyDesc)m_MarkerBodyList[lbId];
			        // calculate the min and max length of the bar according to the min and max
			        // distance to the tracking camera (nearclip/farclip)
			        // get max bar length
                    float focalLengthMM = m_cameraDesc.focalLengthX*m_cameraDesc.pixelSize;
			        float maxlen = focalLengthMM 
				        * element.BarLength / element.NearClip;
			        // get min bar length
			        float minlen = focalLengthMM
				        * element.BarLength / element.FarClip;

			        if (((element.IdMarker-tolerance) <= markerCode) 
				        && (markerCode <= (element.IdMarker+tolerance))
				        && (islen < maxlen))
			        {
				        // now finally a valid line body is found
	                    // set as recognized
	                    element.BarRecognized = true;
				        element.IdPoint = imageMarker[onLinePoint];
                        element.StartLinePoint = imageMarker[startLinePoint];
                    	element.EndLinePoint = imageMarker[endLinePoint];

				        m_recognizedElements.Add( lbId );

				        // delete the three points from the marker vector
				        Vector2 del1 = a;
				        Vector2 del2 = b;
				        Vector2 del3 = c;
                        m_markerVector.Remove(a);
                        m_markerVector.Remove(b);
                        m_markerVector.Remove(c);
				        return true;
			        }
		        }
		        return false;

	        }
	        else
	        {
		        return false;
	        }
        }
        #endregion

        #region Find 5 marker body
        public void FindBody( int barId )
        {
	        TCamMarkerBodyDesc element = (TCamMarkerBodyDesc)m_MarkerBodyList[barId];

	        Vector2 start = element.StartLinePoint;
	        Vector2 end = element.EndLinePoint;
	        Vector2 middle = start + element.CrossPoint*(end-start);
	        Vector2 upRight = new Vector2();
	        Vector2 lowLeft = new Vector2();
            //Vector2 upRightRot;
            //Vector2 lowLeftRot;
	        bool foundBody = false;
	        element.MarkerRecognized = false;

	        // calculate the min and max length of the bar according to the min and max
	        // distance to the tracking camera (nearclip/farclip)
	        // get max bar length
            float focalLengthMM = m_cameraDesc.focalLengthX*m_cameraDesc.pixelSize;
	        float maxlen = focalLengthMM 
		        * element.BarLength / element.NearClip;
	        // get min bar length
	        float minlen = focalLengthMM
		        * element.BarLength / element.FarClip;

	        Vector2 barLine = start - end;
	        barLine.Normalize();

	        // crossing line from the side makers
            //float lineTolerance = 0.2f;

	        if (m_remainingPoints.Count > 1)
	        {
		        for (int i=0; i<m_remainingPoints.Count-1; i++)
		        {
			        for (int j=i+1;j<m_remainingPoints.Count; j++)
			        {
				        Vector2 testPoint = m_remainingPoints[i];
				        Vector2 testLine = m_remainingPoints[j] - m_remainingPoints[i];
				        Vector2 testMiddle = m_remainingPoints[i] + 0.5f*(m_remainingPoints[j] - m_remainingPoints[i]);

				        // for a valid marker the line of the additional two points has to cross the the bar line
				        // the crossing point has to be between both lines

				        // intersection of two lines
				        // line1: a + (lambda)b = x
				        // line2: r + (mue)s = x
				        // when setten line1 = line2 lambda and mue has to be [0..1]
				        Vector2 a = m_remainingPoints[i];
				        Vector2 b = m_remainingPoints[j] - m_remainingPoints[i];
				        Vector2 r = start;
				        Vector2 s = end - start;
				        float lambda, mue;

				        float tempDiv = s.Y*b.X - s.X*b.Y;
				        if (tempDiv != 0)
				        {
					        mue = r.X*b.Y + a.Y*b.X - a.X*b.Y - r.Y*b.X;
					        mue /= tempDiv;
				        }
				        else
				        {
					        mue = -1;
				        }
				        lambda = mue;
				        // now do the same with exchanged lines
				        r = m_remainingPoints[i];
				        s = m_remainingPoints[j] - m_remainingPoints[i];
				        a = start;
				        b = end - start;
				        tempDiv = s.Y*b.X - s.X*b.Y;
				        if (tempDiv != 0)
				        {
					        mue = r.X*b.Y + a.Y*b.X - a.X*b.Y - r.Y*b.X;
					        mue /= tempDiv;
				        }
				        else
				        {
					        mue = -1;
				        }

				        if ((mue>0) && (mue<1) && (lambda>0) && (lambda<1))
				        {
					        float af =(float)( Math.PI + Math.Atan2( (start-testMiddle).Y, (start-testMiddle).X ));
					        float a_right = (float)(Math.PI + Math.Atan2( (m_remainingPoints[i]-testMiddle).Y, (m_remainingPoints[i]-testMiddle).X ));
					        float a_left = (float)(Math.PI + Math.Atan2( (m_remainingPoints[j]-testMiddle).Y, (m_remainingPoints[j]-testMiddle).X ));
					        //float dot = D3DXVec2Dot(&(start-testMiddle), &(*m_remainingPoints[i]-testMiddle) );
					        if (af >= Math.PI)
					        {
						        // test if left really is left
						        if ((a_left < af) && (a_left > af-Math.PI))
						        {
							        upRight = m_remainingPoints[i];
							        lowLeft = m_remainingPoints[j];
						        }
						        // otherwise switch
						        else
						        {
							        upRight = m_remainingPoints[j];
							        lowLeft = m_remainingPoints[i];
						        }
					        }
					        else
					        {
						        // test if right really is right
						        if ((a_right > af) && (a_right < af+Math.PI))
						        {
							        upRight = m_remainingPoints[i];
							        lowLeft = m_remainingPoints[j];
						        }
						        // otherwise switch
						        else
						        {
							        upRight = m_remainingPoints[j];
							        lowLeft = m_remainingPoints[i];
						        }

					        }
					        // test if the middle test point is about in the middle of the points
					        //float diff = D3DXVec2Length( &((upRight-middle)+(lowLeft-middle)) );
					        //if (abs(diff) < (2*lineTolerance))
					        {
						        // test is about the lenght of the line is valid

						        /* old way 
						        float vecA = D3DXVec2Length( &(upRight-middle) );
						        float vecB = D3DXVec2Length( &(lowLeft-middle) );
						        float barlen = D3DXVec2Length( &(start - end) );

						        if ((abs(vecA)+abs(vecB)) < (barlen * 1.25))
						        */
						        // new way according to near and far clip
						        float testLineLen = testLine.Length();
						        if ((testLineLen < maxlen))
						        {
							        foundBody = true;
							        // exit the slopes
							        i = j = (int)m_remainingPoints.Count;
						        }
					        }
				        }
			        }
		        }
	        }
	        if ( foundBody )
	        {

		        element.MarkerRecognized = true;
                if (m_recognizedImagePoints.ContainsKey(m_MarkerBodyList[barId].Id))
                {
                    // error same id recognized both, means one is wrong
                    // to be save delete both (skip)
                    m_recognizedImagePoints.Remove(m_MarkerBodyList[barId].Id);
                }
                else
                {
                    // write found points into dictonary
                    Vector2[] recImagePoints = new Vector2[5];
                    recImagePoints[0] = element.StartLinePoint;
                    recImagePoints[1] = upRight;
                    recImagePoints[2] = element.EndLinePoint;
                    recImagePoints[3] = lowLeft;
                    recImagePoints[4] = element.IdPoint;
                    m_recognizedImagePoints.Add(m_MarkerBodyList[barId].Id, recImagePoints);
                }
            }
        }
        #endregion

     }
}
