﻿/*
 * Copyright (C): Timo Fleisch
 *                2009 Berlin, Germany
 *                http://www.vrhome.de
 *                All Rights Reserved.
 *
 * THIS SOURCECODE IS PUBLISHED UNDER RESTRICTIONS.
 * No unauthorized copy or distribution allowed.
 * Please see website for detailed liscence information.
 * 
 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using WiimoteLib;
using Tgex;
using Microsoft.Xna.Framework;

namespace Tvrx
{
    public class WiiMoteTracker : IMarkerTracker
    {
        // marker order is: left, right, up down
        // create a new instance of the Wiimote
        static WiimoteCollection m_wiimoteCollection  = new WiimoteCollection();
        static int m_wiimoteCount = 0;
        Wiimote wm;
        int m_wiimoteId;
        IPoseEstimate m_poseEstimate;
        MarkerBodyDesc m_markerBodyDesc;
        Vector2 m_principalPoint;
        Vector2 m_resolution;
        float m_pixelSize;
        Vector2[] m_imagePoints = new Vector2[4];
        AverageFilter[] m_imagePointsFilter = new AverageFilter[8];
        Vector2 m_lineStartPoint;
        Vector2 m_lineEndPoint;
        Vector2 m_onLinePoint;
        Vector2 m_freePoint;
        float m_pointLineDist;

        bool m_isTracking = false;
        MarkerTrackerDesc m_markerTrackerDesc;

        // to swap camera axis
        int[] assignAxis = new int[3];
        int[] assignAxisSign = new int[3];
        private int Axis2Number(String s)
        {
            s = s.ToUpper();
            switch (s)
            {
                case "X":
                    return 0;
                case "Y":
                    return 1;
                case "Z":
                    return 2;
            }
            return 0;
        }
        AverageFilter[] m_acceleratorFilter = new AverageFilter[3];
        protected Vector3 m_Rotation;
        /// <summary>
        /// Rotation of wiimote accoring to accelerator values.
        /// </summary>
        public Vector3 Rotation
        {
            get { return m_Rotation; }
            set { m_Rotation = value; }
        }

        public void Initialize(MarkerTrackerDesc desc)
        {
            m_markerTrackerDesc = desc;
            // assign axis
            // axis sign position - if there is a minus sign, 
            // we have to read the next char as axis
            int assp = 0;
            if (desc.XAxis.Length > 0)
            {
                if (desc.XAxis.Substring(0, 1) == "-")
                {
                    assignAxisSign[0] = -1;
                    assp++;
                }
                else
                {
                    assignAxisSign[0] = 1;
                }
                assignAxis[0] = Axis2Number(desc.XAxis.Substring(assp, 1));
            }
            assp = 0;
            if (desc.YAxis.Length > 0)
            {
                if (desc.YAxis.Substring(0, 1) == "-")
                {
                    assignAxisSign[1] = -1;
                    assp++;
                }
                else
                {
                    assignAxisSign[1] = 1;
                }
                assignAxis[1] = Axis2Number(desc.YAxis.Substring(assp, 1));
            }
            assp = 0;
            if (desc.ZAxis.Length > 0)
            {
                if (desc.ZAxis.Substring(0, 1) == "-")
                {
                    assignAxisSign[2] = -1;
                    assp++;
                }
                else
                {
                    assignAxisSign[2] = 1;
                }
                assignAxis[2] = Axis2Number(desc.ZAxis.Substring(assp, 1));
            }

            Initialize();
        }
        public void Initialize()
        {
            // test static variable for first time call
            if (m_wiimoteCount == 0)
            {
                m_wiimoteCollection.FindAllWiimotes();
            }
            if (m_wiimoteCollection.Count <= m_wiimoteCount)
            {
                ErrorHandler.Report("Invalid WiimoteTracker count, only "
                    + m_wiimoteCollection.Count.ToString() + " Wiimotes found");
                return;
            }
            wm = m_wiimoteCollection.ElementAt(m_wiimoteCount);
            m_wiimoteId = m_wiimoteCount;
            m_wiimoteCount++;
            //wm = new Wiimote();
            // setup the event to handle state changes
            wm.WiimoteChanged += wm_WiimoteChanged;

            // setup the event to handle insertion/removal of extensions
            wm.WiimoteExtensionChanged += wm_WiimoteExtensionChanged;
            // create filter for accelerator values
            AverageFilterDesc filterDesc = new AverageFilterDesc();
            filterDesc.numOfValues = 1000;
            for (int i = 0; i < 3; i++)
            {
                m_acceleratorFilter[i] = new AverageFilter();
                m_acceleratorFilter[i].SetFilterDesc(filterDesc);
            }
            // create filter for image points
            filterDesc = new AverageFilterDesc();
            filterDesc.numOfValues = 5;
            for (int i = 0; i < 8; i++)
            {
                m_imagePointsFilter[i] = new AverageFilter();
                m_imagePointsFilter[i].SetFilterDesc(filterDesc);
            }
        }
        public void StartTracking()
        {
            if (!m_isTracking)
            {
                m_isTracking = true;
                // connect to the Wiimote
                try
                {
                    wm.Connect();
                    // set the report type to return the IR sensor and accelerometer data (buttons always come back)
                    wm.SetReportType(InputReport.IRAccel, true);
                    wm.SetLEDs(m_wiimoteId);
                }
                catch
                {
                    ErrorHandler.Report("Cannot connect to Wiimote");
                    m_isTracking = false;
                }
            }
        }
        public void StopTracking()
        {
            if (m_isTracking)
            {
                m_isTracking = false;
                wm.Disconnect();
            }
        }

        public void RegisterMarkerBody(MarkerBodyDesc markerDesc, IPoseEstimate poseEstimate)
        {
            m_poseEstimate = poseEstimate;
            m_markerBodyDesc = markerDesc;
	        // wiimote focal length (meassured)
            float focalLength = 1380;
            // assume 1/4" cd sensor (even though it probably is not)
            m_pixelSize = 0.0035f;
            m_resolution = new Vector2(1024, 768);
            // wiimote center (approx.)
            //m_principalPoint = new Vector2(508, 380);
            m_principalPoint = new Vector2(512, 384);
            focalLength *= m_pixelSize;

            m_poseEstimate.InitializeCameraParameter(focalLength, false, 0.001f, assignAxis, assignAxisSign);
        }

        void wm_WiimoteExtensionChanged(object sender, WiimoteExtensionChangedEventArgs args)
        {
            if (args.Inserted)
                wm.SetReportType(InputReport.IRExtensionAccel, true);    // return extension data
            else
                wm.SetReportType(InputReport.IRAccel, true);            // back to original mode
        }

        List<Vector2> irList = new List<Vector2>();
        void wm_WiimoteChanged(object sender, WiimoteChangedEventArgs args)
        {
            // current state information
            WiimoteState ws = args.WiimoteState;

            if (ws.IRState.IRSensors[0].Found
                && ws.IRState.IRSensors[1].Found
                && ws.IRState.IRSensors[2].Found
                && ws.IRState.IRSensors[3].Found)
            {                
                // put in list
                irList.Clear();
                for (int i = 0; i < 4; i++) irList.Add(
                        new Vector2((float)(ws.IRState.IRSensors[i].Position.X * m_resolution.X) - m_principalPoint.X,
                            (float)(ws.IRState.IRSensors[i].Position.Y * m_resolution.Y) - m_principalPoint.Y));

                m_pointLineDist = float.MaxValue;
                // write all test cases explicite
                // in the end the three point line should be found
                TestPoints(irList[0], irList[1], irList[2], irList[3]);
                TestPoints(irList[0], irList[1], irList[3], irList[2]);
                TestPoints(irList[0], irList[2], irList[1], irList[3]);
                TestPoints(irList[0], irList[2], irList[3], irList[1]);
                TestPoints(irList[0], irList[3], irList[1], irList[2]);
                TestPoints(irList[0], irList[3], irList[2], irList[1]);
                TestPoints(irList[1], irList[2], irList[0], irList[3]);
                TestPoints(irList[1], irList[2], irList[3], irList[0]);
                TestPoints(irList[1], irList[3], irList[0], irList[2]);
                TestPoints(irList[1], irList[3], irList[2], irList[0]);
                TestPoints(irList[2], irList[3], irList[0], irList[1]);
                TestPoints(irList[2], irList[3], irList[1], irList[0]);
                // only remaing test is check if line start and end point is in right order
                //http://www.geocities.com/siliconvalley/2151/math2d.html
                // check start and end line point with free point
                // if clockwise direction then ok, if counterclockwise then exchange start and end point
                Vector2 E1 = m_lineStartPoint - m_freePoint; // P1-P2 
                Vector2 E2 = m_lineEndPoint - m_freePoint; // P3-P2 
                bool clockwise;
                if ((E1.X * E2.Y - E1.Y * E2.X) >= 0) clockwise = true;
                else clockwise = false;
                if (!clockwise)
                {
                    Vector2 tmp = m_lineEndPoint;
                    m_lineEndPoint = m_lineStartPoint;
                    m_lineStartPoint = tmp;
                }
                // now write resulting order to image points
                m_imagePoints[0].X = m_imagePointsFilter[0].Filter(m_lineStartPoint.X);
                m_imagePoints[0].Y = m_imagePointsFilter[1].Filter(m_lineStartPoint.Y);
                m_imagePoints[1].X = m_imagePointsFilter[2].Filter(m_onLinePoint.X);
                m_imagePoints[1].Y = m_imagePointsFilter[3].Filter(m_onLinePoint.Y);
                m_imagePoints[2].X = m_imagePointsFilter[4].Filter(m_lineEndPoint.X);
                m_imagePoints[2].Y = m_imagePointsFilter[5].Filter(m_lineEndPoint.Y);
                m_imagePoints[3].X = m_imagePointsFilter[6].Filter(m_freePoint.X);
                m_imagePoints[3].Y = m_imagePointsFilter[7].Filter(m_freePoint.Y);

                for (int i = 0; i < 4; i++)
                {
                    m_imagePoints[i] *= m_pixelSize;
                }
                
                // send points to estimation
                m_poseEstimate.UpdateImagePoints(m_imagePoints);
            }
            // get accelerator values
            float xAcc = m_acceleratorFilter[0].Filter(ws.AccelState.Values.X);
            float yAcc = m_acceleratorFilter[1].Filter(ws.AccelState.Values.Y);
            float zAcc = m_acceleratorFilter[2].Filter(ws.AccelState.Values.Z);

            float xAxis = (float)(Math.Atan(xAcc / Math.Sqrt(yAcc * yAcc + zAcc * zAcc)));
            float yAxis = (float)(Math.Atan(yAcc / Math.Sqrt(xAcc * xAcc + zAcc * zAcc)));
            float zAxis = (float)(Math.Atan(zAcc / Math.Sqrt(yAcc * yAcc + xAcc * xAcc)));

            // swap rotaiton axis according to config file definition - same as with translation before
            //for (int i = 0; i < 3; i++) swapAxis[i] = angle[i];
            //for (int i = 0; i < 3; i++) angle[i] = m_assignAxisSign[i] * swapAxis[m_assignAxis[i]];
            //float[] origAxis = new float[3];
            // normal position (wiimote at screen)
            if ((assignAxis[0] == 0) && (assignAxis[1] == 1) && (assignAxis[2] == 2))
            {
                m_Rotation.X = yAxis;
                m_Rotation.Z = xAxis;
            }
                // wimmote hanging from ceiling
            else if ((assignAxis[0] == 0) && (assignAxis[1] == 2) && (assignAxis[2] == 1))
            {
                m_Rotation.X = -zAxis;
                m_Rotation.Z = xAxis;
            }
            // for all other wiimote orientation do not calculate angel

        }
        void TestPoints(Vector2 lineStartPoint,
                        Vector2 lineEndPoint,
                        Vector2 onLinePoint,
                        Vector2 freePoint)
        {
            float lambda;
            float dist =
                onLinePoint.DistanceToLine(lineStartPoint, lineEndPoint, out lambda);
            // check if projected point is between line end and start point
            if ((lambda > 0) && (lambda < 1))
            {
                // if distance is short, make this combination the result
                if (dist < m_pointLineDist)
                {
                    m_pointLineDist = dist;
                    m_lineStartPoint = lineStartPoint;
                    m_lineEndPoint = lineEndPoint;
                    m_onLinePoint = onLinePoint;
                    m_freePoint = freePoint;
                }
            }
        }
    }
}
