﻿/*
 * Copyright (C): Timo Fleisch
 *                2009 Berlin, Germany
 *                http://www.vrhome.de
 *                All Rights Reserved.
 *
 * THIS SOURCECODE IS PUBLISHED UNDER RESTRICTIONS.
 * No unauthorized copy or distribution allowed.
 * Please see website for detailed liscence information.
 * 
 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;

namespace Tvrx
{
    public class TrackedDevice
    {
        protected IPoseEstimate m_poseEstimate;

        #region transforms
        protected Matrix m_TrackerWorldTransform = Matrix.Identity;
        /// <summary>
        /// Transformation of the tracker sensor device e.g. camera.
        /// </summary>
        public Matrix TrackerWorldTransform
        {
            get { return m_TrackerWorldTransform; }
            set { m_TrackerWorldTransform = value; }
        }
        protected Matrix m_DeviceWorldTransform = Matrix.Identity;
        /// <summary>
        /// Transforms the whole device in world space.
        /// </summary>
        public Matrix DeviceWorldTransform
        {
            get { return m_DeviceWorldTransform; }
            set { m_DeviceWorldTransform = value; }
        }
        protected Matrix m_DeviceLocalTransform = Matrix.Identity;
        /// <summary>
        /// Transforms the device itself according to its local origin e.g. zero marker.
        /// </summary>
        public Matrix DeviceLocalTransform
        {
            get { return m_DeviceLocalTransform; }
            set { m_DeviceLocalTransform = value; }
        }

        #endregion

        protected bool m_isTracking = false;

        protected Vector3 m_rawRotation;
        protected Vector3 m_rawTranslation;
        protected IDataFilter[] m_translationFilter = new IDataFilter[3];
        protected IDataFilter[] m_rotationFilter = new IDataFilter[3];
        public void SetTranslationFilter(IDataFilter x, IDataFilter y, IDataFilter z)
        {
            m_translationFilter[0] = x;
            m_translationFilter[1] = y;
            m_translationFilter[2] = z;
        }
        public void SetRotationFilter(IDataFilter x, IDataFilter y, IDataFilter z)
        {
            m_rotationFilter[0] = x;
            m_rotationFilter[1] = y;
            m_rotationFilter[2] = z;
        }
        // reverse direction of axis if bool value is true
        protected bool[] m_reverseTranslation = new bool[3];
        protected bool[] m_reverseRotation = new bool[3];
        public void SetReverseAxis(bool[] reverseTranslation, bool[] reverseRotation)
        {
            for (int i = 0; i<3; i++)
            {
                m_reverseTranslation[i] = reverseTranslation[i];
                m_reverseRotation[i] = reverseRotation[i];
            }
        }

        protected Vector3 m_Translation;
        /// <summary>
        /// Translation of device.
        /// </summary>
        public Vector3 Translation
        {
            get { return m_Translation; }
            set { m_Translation = value; }
        }
        protected Quaternion m_Rotation;
        /// <summary>
        /// Rotation of device in euler angles.
        /// </summary>
        public Quaternion Rotation
        {
            get { return m_Rotation; }
            set { m_Rotation = value; }
        }

        protected int m_Id;
        /// <summary>
        /// Id of tracked device
        /// </summary>
        public int Id
        {
            get { return m_Id; }
            set { m_Id = value; }
        }

        public virtual void Initialize() { }
        /// <summary>
        /// This starts a new thread that continously tracks the device
        /// </summary>
        public virtual void StartTracking() 
        {
            m_isTracking = true;
        }
        public virtual void StopTracking() 
        {
            m_isTracking = false;
        }

        public virtual void Filter() 
        {
            m_rawTranslation.X = m_translationFilter[0].Filter(m_rawTranslation.X);
            m_rawTranslation.Y = m_translationFilter[1].Filter(m_rawTranslation.Y);
            m_rawTranslation.Z = m_translationFilter[2].Filter(m_rawTranslation.Z);
            m_rawRotation.X = m_rotationFilter[0].Filter(m_rawRotation.X);
            m_rawRotation.Y = m_rotationFilter[1].Filter(m_rawRotation.Y);
            m_rawRotation.Z = m_rotationFilter[2].Filter(m_rawRotation.Z);
        }
        /// <summary>
        /// Usually tracked device will have its data realtive to the tracking
        /// camera coordinate system.
        /// Use this function to get in the right place in the 3d application.
        /// </summary>
        public virtual void TransformToVirtualSpace() 
        {
            Matrix bodyTransformMatrix =
                Matrix.CreateFromYawPitchRoll(m_rawRotation.Y, m_rawRotation.X, m_rawRotation.Z)
                * Matrix.CreateTranslation(m_rawTranslation);

            Matrix result = m_TrackerWorldTransform * m_DeviceWorldTransform;
            result = bodyTransformMatrix * result;
            result = m_DeviceLocalTransform * result;

            Vector3 scale;
            result.Decompose(out scale, out m_Rotation, out m_Translation);
        }
        public void Update()
        {
            if (m_isTracking)
            {
                m_poseEstimate.GetTransform(out m_rawTranslation, out m_rawRotation);
                if (m_reverseTranslation[0]) m_rawTranslation.X *= -1;
                if (m_reverseTranslation[1]) m_rawTranslation.Y *= -1;
                if (m_reverseTranslation[2]) m_rawTranslation.Z *= -1;
                if (m_reverseRotation[0]) m_rawRotation.X *= -1;
                if (m_reverseRotation[1]) m_rawRotation.Y *= -1;
                if (m_reverseRotation[2]) m_rawRotation.Z *= -1;
                Filter();
                TransformToVirtualSpace();
            }
        }
    }
}
