﻿/*
 * Copyright (C): Timo Fleisch
 *                2009 Berlin, Germany
 *                http://www.vrhome.de
 *                All Rights Reserved.
 *
 * THIS SOURCECODE IS PUBLISHED UNDER RESTRICTIONS.
 * No unauthorized copy or distribution allowed.
 * Please see website for detailed liscence information.
 * 
 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;

namespace Tvrx
{
    public class TrackerManager
    {
        protected List<TrackedDevice> m_TrackedDeviceList = new List<TrackedDevice>();
        /// <summary>
        /// Help
        /// </summary>
        public List<TrackedDevice> TrackedDeviceList
        {
            get { return m_TrackedDeviceList; }
            set { m_TrackedDeviceList = value; }
        }
        protected List<TrackerProxy> m_TrackerProxies = new List<TrackerProxy>();
        /// <summary>
        /// List of TrackerProxies. This is the main interface for the application.
        /// </summary>
        public List<TrackerProxy> TrackerProxies
        {
            get { return m_TrackerProxies; }
            set { m_TrackerProxies = value; }
        }

        #region singelton
        // create singleton
        static readonly TrackerManager instance = new TrackerManager();
        /// <summary>
        /// Get instance of singleton.
        /// </summary>
        public static TrackerManager Instance
        {
            get
            {
                return instance;
            }
        }
        /// <summary>
        /// Explicit static constructor to tell C# compiler
        /// not to mark type as beforefieldinit
        /// </summary>
        static TrackerManager()
        {
        }
        #endregion
        /// <summary>
        /// Constructor
        /// </summary>
        TrackerManager()
        {
        }

        public void Initialize()
        {
            List<MarkerBodyDesc> markerDescList;
            List<MarkerTrackerDesc> markerTrackerDescList;
            Dictionary<int,IMarkerTracker> markerTrackerList = new Dictionary<int,IMarkerTracker>();
            List<TrackerProxyDesc> trackerProxyDescList;
            List<TrackedDeviceDesc> trackedDeviceDescList;
            List<DataFilterDesc> dataFilterDescList;

            ConfigReader.ReadTrackerConfig(
                out markerDescList, 
                out markerTrackerDescList, 
                out trackedDeviceDescList,
                out trackerProxyDescList,
                out dataFilterDescList);

            // create marker tracker (cameras)
            foreach (MarkerTrackerDesc mtd in markerTrackerDescList)
            {
                if (mtd.CameraModel == "Wiimote")
                {
                    WiiMoteTracker wiimoteTracker = new WiiMoteTracker();
                    wiimoteTracker.Initialize(mtd);
                    markerTrackerList.Add(mtd.Id, wiimoteTracker);
                }
                else
                {
                    TCamTracker tcam = new TCamTracker();
                    tcam.Initialize(mtd);
                    markerTrackerList.Add(mtd.Id, tcam);
                }
            }
            // create tracked devices
            foreach (TrackedDeviceDesc tdd in trackedDeviceDescList)
            {
                TrackedDevice trackedDevice = null;
                if (tdd is TrackedMarkerDesc)
                {
                    TrackedMarker trackedMarker = new TrackedMarker();
                    trackedDevice = trackedMarker;
                    TrackedMarkerDesc tmd = (TrackedMarkerDesc)tdd;
                    TCamMarkerBodyDesc markerBodyDesc = 
                        (TCamMarkerBodyDesc)markerDescList.Find(desc => desc.Id == tmd.MarkerBodyId);
                    trackedMarker.Initialize();
                    TCamTracker tcam = (TCamTracker)markerTrackerList[tdd.TrackerCamId];
                    trackedMarker.SetTracker(tcam);
                    trackedMarker.SetMarkerBody(markerBodyDesc);
                }
                else if (tdd is TrackedWiiMoteDesc)
                {
                    TrackedWiiMote trackedWiiMote = new TrackedWiiMote();
                    trackedDevice = trackedWiiMote;
                    TrackedWiiMoteDesc twd = (TrackedWiiMoteDesc)tdd;
                    MarkerBodyDesc markerBodyDesc =
                        markerDescList.Find(desc => desc.Id == twd.MarkerBodyId);
                    trackedWiiMote.Initialize();
                    WiiMoteTracker wiiTracker = (WiiMoteTracker)markerTrackerList[tdd.TrackerCamId];
                    trackedWiiMote.SetTracker(wiiTracker);
                    trackedWiiMote.SetMarkerBody(markerBodyDesc);
                }
                MarkerTrackerDesc trackerDesc =
                    markerTrackerDescList.Find(m => m.Id == tdd.TrackerCamId);
                trackedDevice.Id = tdd.Id;
                trackedDevice.SetReverseAxis(tdd.ReverseTranslation, tdd.ReverseRotation);
                trackedDevice.DeviceLocalTransform =
                    Matrix.CreateFromYawPitchRoll(
                        tdd.LocalRotation.Y, tdd.LocalRotation.X, tdd.LocalRotation.Z)
                    * Matrix.CreateTranslation(tdd.LocalTranslation * trackerDesc.Scale);
                trackedDevice.DeviceWorldTransform =
                    Matrix.CreateFromYawPitchRoll(tdd.WorldRotation.Y, tdd.WorldRotation.X, tdd.WorldRotation.Z)
                    * Matrix.CreateTranslation(tdd.WorldTranslation * trackerDesc.Scale);
                trackedDevice.TrackerWorldTransform =
                    Matrix.CreateFromYawPitchRoll(
                        trackerDesc.Rotation.Y, trackerDesc.Rotation.X, trackerDesc.Rotation.Z)
                    * Matrix.CreateTranslation(trackerDesc.Translation * trackerDesc.Scale);
                // create data filter
                // translation filter
                DataFilterDesc filterDesc = dataFilterDescList.Find(d => d.id == tdd.TranslationFilterId);
                Type t = Type.GetType("Tvrx." + filterDesc.className);
                if (t != null)
                {
                    IDataFilter[] filter = new IDataFilter[3];
                    for (int i = 0; i < 3; i++)
                    {
                        filter[i] = (IDataFilter)Activator.CreateInstance(t);
                        filter[i].SetFilterDesc(filterDesc);
                    }
                    trackedDevice.SetTranslationFilter(filter[0], filter[1], filter[2]);
                }
                // rotation filter
                filterDesc = dataFilterDescList.Find(d => d.id == tdd.RotationFilterId);
                t = Type.GetType("Tvrx." + filterDesc.className);
                if (t != null)
                {
                    IDataFilter[] filter = new IDataFilter[3];
                    for (int i = 0; i < 3; i++)
                    {
                        filter[i] = (IDataFilter)Activator.CreateInstance(t);
                        filter[i].SetFilterDesc(filterDesc);
                    }
                    trackedDevice.SetRotationFilter(filter[0], filter[1], filter[2]);
                }
                // add tracked device to list
                m_TrackedDeviceList.Add(trackedDevice);
            }
            // create tracker proxies
            foreach (TrackerProxyDesc tpd in trackerProxyDescList)
            {
                TrackerProxy trackerProxy = new TrackerProxy();
                trackerProxy.Id = tpd.Id;
                // add translation devices
                foreach (int translationId in tpd.TranslationDeviceIds)
                {
                    trackerProxy.TranslationDevices.Add(m_TrackedDeviceList.Find(d => d.Id == translationId));
                }
                // add rotation devices
                foreach (int rotationId in tpd.RotationDeviceIds)
                {
                    trackerProxy.RotationDevices.Add(m_TrackedDeviceList.Find(d => d.Id == rotationId));
                }
                m_TrackerProxies.Add(trackerProxy);
            }

        }

        public void StartTracking()
        {
            TrackedDeviceList.ForEach(device => device.StartTracking());
        }

        public void StopTracking()
        {
            TrackedDeviceList.ForEach(device => device.StopTracking());
        }

        public void Update()
        {
            TrackedDeviceList.ForEach(device => device.Update());
        }

        public Matrix GetDeviceTransform(int deviceId)
        {
            TrackedDevice trackedDevice = m_TrackedDeviceList.Find(device => device.Id == deviceId);
            return Matrix.CreateFromQuaternion(trackedDevice.Rotation)
                * Matrix.CreateTranslation(trackedDevice.Translation);
        }

        public Vector3 GetDeviceTranslation(int deviceId)
        {
            TrackedDevice trackedDevice = m_TrackedDeviceList.Find(device => device.Id == deviceId);
            return trackedDevice.Translation;
        }
        public Matrix GetProxyTransform(int proxyId)
        {
            return m_TrackerProxies.Find(p => p.Id == proxyId).GetTransform();
        }
    }
}
